/*
 * Copyright(c) 1997 by Jun-ichiro Itoh. All rights reserved.
 * Freely redistributable unless otherwise noted.  Absolutely no warranty.
 *
 * author contact: Jun-ichiro itojun Itoh <itojun@mt.cs.keio.ac.jp>
 * $Id: main.c,v 1.1.1.1 1998/02/23 14:22:24 itojun Exp $
 */
#include "common.h"
#include "crctab.c"

#if HAVE_TERMIOS_H
# include <termios.h>
#else
/*# if HAVE_TERMIO_H*/
# if 0
#  include <termio.h>
# else
#  if HAVE_SYS_IOCTL_H
#   include <sys/ioctl.h>
#  endif
#  include <sgtty.h>
# endif
#endif
#include <setjmp.h>
#include <sys/stat.h>

/* serial lines */
static int baudrateflag = 2400;
static char *n0;
static int fd0;

static int dumpflag = 0;
static int dumpmaxlen = 20;
static int debugflag = 0;
static int openflag = O_WRONLY | O_TRUNC | O_CREAT;
static int verbose = 0;
static int echobackrate = 5;

static enum {
	NOTHING, RESET, SHOWNPICTS, GETPICTS, GETPICT1, GETSTATUS,
} mode = SHOWNPICTS;

static int startpict = 0;
static int endpict = 0;
static int allpict = 0;

static char *nametemplate = "r03%05d.jpg";

#define dprintf(x)						\
	{ if (debugflag) {					\
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);	\
		fprintf x;					\
	  }							\
	}

/* uid */
static uid_t uid, euid;
static gid_t gid, egid;
static int uidswapped = 0;

/* uucp locking */
static int useuucplock = 1;
extern int uu_lock __P((char *));
extern int uu_unlock __P((char *));

/* error handling */
jmp_buf fatal;

void
daemonuid()
{
	if (uidswapped == 0)
		return;

#ifdef HAVE_SETREUID
	setreuid(uid, euid);
	setregid(gid, egid);
#else
	setuid(uid);
	seteuid(euid);
	setgid(gid);
	setegid(egid);
#endif
	uidswapped = 0;
}

void
useruid()
{
	if (uidswapped == 1)
		return;

#ifdef HAVE_SETREUID
	setregid(egid, gid);
	setreuid(euid, uid);
#else
	setgid(egid);
	setegid(gid);
	setuid(euid);
	seteuid(uid);
#endif
	uidswapped = 1;
}

static int
uucplock(name)
	char *name;
{
	int ret;

	if (useuucplock) {
		daemonuid();
		ret = uu_lock(rindex(name, '/') + 1);
		useruid();
	} else
		ret = 0;
	return ret;
}

static int
uucpunlock(name)
	char *name;
{
	int ret;

	if (useuucplock) {
		daemonuid();
		ret = uu_unlock(rindex(name, '/') + 1);
		useruid();
	} else
		ret = 0;
	return ret;
}

static char *
canondev(name)
	char *name;
{
	int fd;
	char *p;

	daemonuid();

	p = malloc(strlen(name) + 1);
	if (!p) {
		perror("malloc");
		goto back;
	}
	strcpy(p, name);
	fd = open(p, O_RDWR|O_NONBLOCK);
	if (0 <= fd) {
		close(fd);
		goto back;
	}

	p = realloc(p, strlen(name) + 6);
	if (!p) {
		perror("realloc");
		goto back;
	}
	strcpy(p, "/dev/");
	strcat(p, name);
	fd = open(p, O_RDWR|O_NONBLOCK);
	if (0 <= fd) {
		close(fd);
		goto back;
	}

	free(p);
	p = NULL;
back:
	useruid();
	return p;
}

speed_t
baudconv(int baud)
{
#define	BAUDCASE(x)	case (x): { ret = B##x; break; }
	speed_t ret;

	ret = (speed_t) baud;
	switch (baud) {
	/* POSIX defined baudrates */
	BAUDCASE(0);	/*is it meaningful? */
	BAUDCASE(50);
	BAUDCASE(75);
	BAUDCASE(110);
	BAUDCASE(134);
	BAUDCASE(150);
	BAUDCASE(200);
	BAUDCASE(300);
	BAUDCASE(600);
	BAUDCASE(1200);
	BAUDCASE(1800);
	BAUDCASE(2400);
	BAUDCASE(4800);
	BAUDCASE(9600);
	BAUDCASE(19200);
	BAUDCASE(38400);

	/* non POSIX values */
#ifdef B7200
	BAUDCASE(7200);
#endif
#ifdef B14400
	BAUDCASE(14400);
#endif
#ifdef B28800
	BAUDCASE(28800);
#endif
#ifdef B57600
	BAUDCASE(57600);
#endif
#ifdef B115200
	BAUDCASE(115200);
#endif
#ifdef B230400
	BAUDCASE(230400);
#endif

	/* last resort */
	default:
		fprintf(stderr, "no defined value for baudrate %d found, "
			"use the value without conversion\n", baud);
	}

	return ret;
#undef BAUDCASE
}

int
setbaud(fd, baud)
	int fd;
	int baud;
{
#if HAVE_TERMIOS_H
	/* termios */
	struct termios tio;

	if (tcgetattr(fd, &tio) < 0) {
		perror("tcgetattr");
		return 1;
	}
	tio.c_iflag = 0;
	tio.c_oflag = 0;
	tio.c_cflag = CS8 | CREAD | CLOCAL;
	tio.c_lflag = 0;
	tio.c_cc[VMIN] = 1;
	tio.c_cc[VTIME] = 5;
	cfsetispeed(&tio, baudconv(baud));
	cfsetospeed(&tio, baudconv(baud));
	if (tcsetattr(fd, TCSANOW, &tio) < 0) {
		perror("tcsetattr");
		return 1;
	}
#else
/*#if HAVE_TERMIO_H*/
# if 0
	/* termio */
# else
	/* sgtty */
	struct sgttyb ttyb;

	if (ioctl(fd, TIOCGETP, &ttyb) < 0) {
		perror("ioctl(TIOCGETP)");
		return 1;
	}
	ttyb.sg_ispeed = baud;
	ttyb.sg_ospeed = baud;
	ttyb.sg_flags = 0;
	if (ioctl(fd, TIOCSETP, &ttyb) < 0) {
		perror("ioctl(TIOCSETP)");
		return 1;
	}
# endif
#endif

	dprintf((stderr, "baudrate set to %d\n", baud));
	return 0;
}

void
dump_stream(dir, buf, len)
	int dir;
	char *buf;
	int len;
{
	size_t i;
	int truncate;

	if (!dumpflag)
		return;

	truncate = 0;
	if (dumpmaxlen < len) {
		len = dumpmaxlen;
		truncate = 1;
	}

	dir &= 0xff;
	if (dir == '>')
		fprintf(stderr, "camera>cpu: ");
	else
		fprintf(stderr, "cpu>camera: ");

	for (i = 0; i < len; i++)
		fprintf(stderr, "%02x ", buf[i] & 0xff);
	if (truncate)
		fprintf(stderr, "...");
	fprintf(stderr, "\n");
}

/*------------------------------------------------------------*/

char ricoh_buf[1024 * 4];
size_t ricoh_len;

void
ricoh_init()
{
	/* initialize buffers */
	ricoh_len = 0;
}

int
ricoh_wait(rlen)
	int rlen;
{
	fd_set rdfd;
	int maxfd;
	int len;
	struct timeval t;

	while (ricoh_len < rlen) {
		/* obtain chars from the camera */
		FD_ZERO(&rdfd);
		FD_SET(fd0, &rdfd);
		maxfd = fd0;
		/* timeout: 2sec */
		t.tv_sec = 2;
		t.tv_usec = 0;
		switch (select(maxfd + 1, &rdfd, NULL, NULL, &t)) {
		case -1:
			if (errno == EINTR)
				break;
			perror("select");
			exit(1);
		case 0:
			dprintf((stderr, "read timeout.\n"));
			return 1;
		default:
			break;
		}

		if (FD_ISSET(fd0, &rdfd)) {
			len = read(fd0, ricoh_buf + ricoh_len,
				sizeof(ricoh_buf) - ricoh_len);
			ricoh_len += len;
		} else {
			dprintf((stderr, "something wrong in ricoh_get\n"));
			return 1;
		}
	}
	dump_stream('>', ricoh_buf, ricoh_len);

	return 0;
}

int
ricoh_get(buf, len)
	u_char *buf;
	int len;
{
	switch (ricoh_wait(len)) {
	case 1:
		dprintf((stderr, "timeouted in ricoh_get\n"));
		fprintf(stderr, "camera not ready.\n");
		longjmp(fatal, 1);
	case 0:
		break;
	}

	if (ricoh_len < len)
		abort();

	memcpy(buf, ricoh_buf, len);
	if (ricoh_len - len)
		memmove(ricoh_buf, ricoh_buf + len, ricoh_len - len);
	ricoh_len -= len;

	return 0;
}

int
ricoh_put(buf, len)
	u_char *buf;
	int len;
{
	int wlen;
	int i;

	/* Thanks goes to Thierry Bousch */
	for (i = 0; i < len; i++) {
		wlen = write(fd0, &buf[i], 1);
		if (wlen != 1) {
			dprintf((stderr, "failed in ricoh_put\n"));
			return 1;
		}
#if HAVE_TERMIOS_H
		/*termios*/
		tcdrain(fd0);
#else
/*#if HAVE_TERMIO_H*/
#if 0
		/*termio - yet to be done*/
#else
		/*sgtty*/
		ioctl(fd0, TCDRAIN, 0);
#endif
#endif
	}

	dump_stream('<', buf, len);
	return 0;
}

/*------------------------------------------------------------*/

int
ricoh_sendcmd(ch, buf, len, blkno)
	u_char ch;
	u_char *buf;
	int len;
	u_char blkno;
{
	u_char tbuf[2];
	u_short crc;
	size_t i;
	int err;

	tcdrain(fd0);

	/* generate crc */
	crc = 0;
	crc = updcrc(ch & 0xff, crc);
	crc = updcrc(len & 0xff, crc);
	for (i = 0; i < len; i++) {
		crc = updcrc(buf[i] & 0xff, crc);
	}

	tbuf[0] = 0x10;
	tbuf[1] = 0x02;
	err = ricoh_put(tbuf, 2);

	tbuf[0] = ch;
	tbuf[1] = len & 0xff;
	err = ricoh_put(tbuf, 2);

	for (i = 0; i < len; i++) {
		/* 0x10 must be escaped */
		if (buf[i] == 0x10)
			err += ricoh_put(&buf[i], 1);
		err += ricoh_put(&buf[i], 1);
	}

	tbuf[0] = 0x10;
	tbuf[1] = 0x03;
	err = ricoh_put(tbuf, 2);

	tbuf[0] = crc & 0x00ff;
	tbuf[1] = (crc & 0xff00) >> 8;
	err += ricoh_put(tbuf, 2);

	tbuf[0] = len + 2;
	tbuf[1] = blkno;
	err += ricoh_put(tbuf, 2);

	return err ? 1 : 0;
}

int
ricoh_getpacket(ch, buf, plen, ack, blk)
	u_char *ch;
	u_char *buf;
	int *plen;
	int *ack;
	u_char *blk;
{
	enum { PRESYNC, SYNC, PACKET, CRC, SKIPCRC } state;
	int gotpacket;
	int gotcrc;
	int gotack;
	int badcrc;

	u_char tbuf[2];
	u_char len;
	int i;

	/* initialize */
	state = PRESYNC;
	gotpacket = gotcrc = gotack = badcrc = 0;
	*plen = 0;
	*ack = 0;
	*blk = 0;

	while (1) {
		switch (state) {
		case PRESYNC:
			if (ricoh_get(tbuf, 1)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"getting sync\n"));
				goto fatal;
			}
			if (tbuf[0] == 0x10)
				state = SYNC;
			else {
				dprintf((stderr, "err in ricoh_getpacket: "
						"bad sync %02x: retry\n",
						tbuf[0]));
			}
			break;

		case SYNC:
			if (ricoh_get(tbuf, 1)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"getting type\n"));
				goto fatal;
			}
			if (tbuf[0] == 0x06) {	/*ACK*/
				*ack = 1;
				goto done;
			} else if (tbuf[0] == 0x15)	/*NACK*/
				goto fatal;
			else if (tbuf[0] == 0x02)
				state = PACKET;
			else if (tbuf[0] == 0x03 || tbuf[0] == 0x17) {
				if (gotpacket)
					state = CRC;
				else
					state = SKIPCRC;
			} else if (tbuf[0] == 0x10) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"dup sync char: syncagain\n"));
				/* keep state as is */
			} else {
				dprintf((stderr, "err in ricoh_getpacket: "
						"bad type %02x: sync again\n",
						tbuf[0]));
				state = PRESYNC;
			}
			break;

		case PACKET:
			if (ricoh_get(ch, 1)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"getting cmdch\n"));
				goto fatal;
			}
			if (ricoh_get(&len, 1)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"getting len\n"));
				goto fatal;
			}
			i = 0;
			while (i < len) {
				if (ricoh_get(&buf[i], 1)) {
					dprintf((stderr,
						"err in ricoh_getpacket: "
						"getting body\n"));
					goto fatal;
				}
				if (buf[i] != 0x10) {
					i++;
					continue;
				}

				/*
				 * 0x10 case: just drop the next byte.
				 * XXX should try to sync again.
				 */
				if (ricoh_get(&buf[i], 1)) {
					dprintf((stderr,
						"err in ricoh_getpacket: "
						"getting body(after 0x10)\n"));
					goto fatal;
				}
				i++;
			}
			gotpacket++;
			state = PRESYNC;
			break;

		case CRC:
			if (ricoh_get(tbuf, 2)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"getting crc\n"));
				goto fatal;
			}

		    {
			u_short crc;
			int i;

			crc = 0;
			crc = updcrc(*ch & 0xff, crc);
			crc = updcrc(len & 0xff, crc);
			for (i = 0; i < len; i++) {
				crc = updcrc(buf[i] & 0xff, crc);
			}

			if ((crc & 0xff) == tbuf[0]
			 && ((crc >> 8) & 0xff) == tbuf[1])
				badcrc = 0;
			else
				badcrc = 1;
			dprintf((stderr, "crc: %04x<->%02x%02x: %s\n",
				crc, tbuf[1], tbuf[0],
				badcrc ? "bad" : "good"));
		    }

			if (ricoh_get(tbuf, 2)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"getting crc tail\n"));
				goto fatal;
			}
			if (tbuf[0] != len + 2) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"bad crc length (%d %d)\n",
						tbuf[0], len + 2));
				badcrc++;
			}
			*blk = tbuf[1];
			gotcrc++;
			state = PRESYNC;
			break;

		case SKIPCRC:
			if (ricoh_get(tbuf, 4)) {
				dprintf((stderr, "err in ricoh_getpacket: "
						"skipping crc\n"));
				goto fatal;
			}
			state = PRESYNC;
			break;
		}

		if (gotpacket && gotcrc) {
			if (!badcrc) {
				tbuf[0] = 0x10;
				tbuf[1] = 0x06;
				ricoh_put(tbuf, 2);
				*plen = len;
				goto done;
			} else {
				tbuf[0] = 0x10;
				tbuf[1] = 0x15;
				ricoh_put(tbuf, 2);
				gotpacket = gotcrc = 0;
				state = PRESYNC;
				/* again */
			}
		}
	}

done:
	return 0;
fatal:
	return 1;
}

/*------------------------------------------------------------*/

int
ricoh_setspeed(baud)
	int baud;
{
	u_char ch;
	u_char buf[10];
	int len;
	int ack;
	u_char blk;
	int err;
	char value;

	err = 0;

	tcdrain(fd0);

	switch (baud) {
	case -1:
	case 2400:
		value = 0x00;
		break;
	case 4800:
		value = 0x01;
		break;
	case 9600:
		value = 0x02;
		break;
	case 19200:
		value = 0x03;
		break;
	case 38400:
		value = 0x04;
		break;
	case 57600:
		value = 0x05;
		break;
	case 115200:
		value = 0x07;
		break;
	default:
		/* unsupported baud rate. */
		dprintf((stderr, "unsupported baudrate %d\n", baud));
		return 1;
	}

	/* set baudrate */
	buf[0] = value;
	ricoh_sendcmd('2', buf, 1, 0x00);

	usleep(20*1000);

	tcdrain(fd0);

	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack);

#if 0
	tcdrain(fd0);
#endif

	usleep(20*1000);

	if (baud == -1)
		err += setbaud(fd0, 2400);
	else
		err += setbaud(fd0, baud);

	usleep(1000*1000);

	return err ? 1 : 0;
}

int
ricoh_hello()
{
	u_char ch;
	u_char buf[1024];
	int len;
	int ack;
	u_char blk;
	int i;
	int err;

	err = 0;

	ricoh_sendcmd(0x31, "\0\0\0", 3, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack || blk);
	if (debugflag) {
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);
		fprintf(stderr, "hello: 31 00 00 00 -> ");
		for (i = 0; i < len; i++)
			fprintf(stderr, "%02x ", buf[i]);
		fprintf(stderr, "\n");
	}

	return err ? 1 : 0;
}

int
ricoh_getnpicts(n)
	int *n;
{
	u_char ch;
	u_char buf[1024];
	int len;
	int ack;
	u_char blk;
	int i;
	int err;

	err = 0;

	ricoh_sendcmd('Q', "\0\001", 2, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack);
	if (debugflag) {
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);
		fprintf(stderr, "getnpicts: Q 00 01: ");
		for (i = 0; i < len; i++)
			fprintf(stderr, "%02x ", buf[i]);
		fprintf(stderr, "\n");
	}

	*n = buf[2];

	return err ? 1 : 0;
}

/*
 * Bunch of unknown commands are sent here.
 */
int
ricoh_magicword()
{
	u_char ch;
	u_char buf[1024];
	int len;
	int ack;
	u_char blk;
	int i;
	int err;

	err = 0;

	ricoh_sendcmd('Q', "\0\0", 2, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack || blk);
	if (debugflag) {
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);
		fprintf(stderr, "Q 00 00: ");
		for (i = 0; i < len; i++)
			fprintf(stderr, "%02x ", buf[i]);
		fprintf(stderr, "\n");
	}

	ricoh_sendcmd('Q', "\022", 1, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack || blk);
	if (debugflag) {
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);
		fprintf(stderr, "Q 12: ");
		for (i = 0; i < len; i++)
			fprintf(stderr, "%02x ", buf[i]);
		fprintf(stderr, "\n");
	}

	ricoh_sendcmd(0x96, "\0", 0, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack || blk);
	if (debugflag) {
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);
		fprintf(stderr, "96: ");
		for (i = 0; i < len; i++)
			fprintf(stderr, "%02x ", buf[i]);
		fprintf(stderr, "\n");
	}

	return err ? 1 : 0;
}

int
ricoh_getpict(n, fd)
	int n;
	int fd;
{
	u_char ch;
	u_char buf[1024 * 4];
	int len;
	int ack;
	u_char blk;
	int i;

	int totlen;
	int flen;
	int err;

	err = 0;

	buf[0] = n & 0xff;
	buf[1] = 0x00;
	ricoh_sendcmd(0xa0, buf, 2, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack);
#if 0
	fprintf(stderr, "get: A0 %02d 00 -> ", n);
	for (i = 0; i < len; i++)
		fprintf(stderr, "%02x ", buf[i]);
	fprintf(stderr, "\n");
#endif

	totlen = (buf[16] << 16) | (buf[15] << 8) | buf[14];
	flen = 0;
	while (flen < totlen) {
		do {
			err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
		} while (ack);

		write(fd, buf, len);
		flen += len;

		if (verbose && (blk % echobackrate == 0)) {
			fprintf(stderr,
				"got block %3d: %d/%d \r",
					blk, flen, totlen);
		}

	}
	if (verbose) {
		fprintf(stderr,
			"got block %3d: %d/%d ...done%s\n",
			blk, flen, totlen, err ? " with err" : "");
	}

	return err ? 1 : 0;
}

int
ricoh_bye()
{
	u_char ch;
	u_char buf[1024];
	int len;
	int ack;
	u_char blk;
	int i;
	int err;

	err = 0;

	ricoh_sendcmd(0x37, "\0", 1, 0x00);
	do {
		err += ricoh_getpacket(&ch, buf, &len, &ack, &blk);
	} while (ack);
	if (debugflag) {
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);
		fprintf(stderr, "bye: 37 00 -> ");
		for (i = 0; i < len; i++)
			fprintf(stderr, "%02x ", buf[i]);
		fprintf(stderr, "\n");
	}

	return err ? 1 : 0;
}

void
mainloop()
{
	int n;
	char fnamebuf[MAXPATHLEN];
	int i;
	int fd;

	switch (setjmp(fatal)) {
	case 0:
		break;
	default:
		goto fatalerr;
	}

	/*initial value*/
	if (setbaud(fd0, 2400)) {
		fprintf(stderr, "can't set baudrate\n");
		goto fatalerr;
	}

	ricoh_init();

	ricoh_hello();

	if (mode == RESET) {
		ricoh_bye();
		return;
	}

	ricoh_setspeed(baudrateflag);

	ricoh_hello();

	ricoh_magicword();

	if (ricoh_getnpicts(&n)) {
		fprintf(stderr, "error getting npicts\n");
		ricoh_bye();
		goto fatalerr;
	}

	/*
	 * parameter clarification
	 */
	if (n) {
		if (allpict) {
			startpict = 1;
			endpict = n;
		} else {
			/* default settings */
			if (startpict == 0)
				startpict = 1;
			if (endpict == 0)
				endpict = n;
		}
		if (startpict < 1 || n < startpict) {
			fprintf(stderr,
				"illegal startpict %d (must be 1 to %d)\n",
				startpict, n);
			ricoh_bye();
			goto fatalerr;
		}
		if (endpict < 1 || n < endpict) {
			fprintf(stderr,
				"illegal endpict %d (must be 1 to %d)\n",
				endpict, n);
			ricoh_bye();
			goto fatalerr;
		}
		if (endpict < startpict) {
			int tmp;
			tmp = startpict;
			startpict = endpict;
			endpict = tmp;
		}
	} else {
		if (allpict || startpict || endpict) {
			fprintf(stderr, "no picts at all\n");
			ricoh_bye();
			goto fatalerr;
		}
	}

	switch (mode) {
	case SHOWNPICTS:
		fprintf(stderr, "picts=%d", n);
#if 0
		if (ricoh_getnpictremain(&n)) {
			fprintf(stderr, "error getting npictremain\n");
			goto fatalerr;
		}
		fprintf(stderr, "/remain=%d", n);
#endif
		fprintf(stderr, "\n");
		break;

	case GETPICTS:
		if (!n) {
			fprintf(stderr, "no picts at all\n");
			break;
		}
		for (i = startpict; i <= endpict; i++) {
			sprintf(fnamebuf, nametemplate, i);
			if (verbose) {
				fprintf(stderr, "getting image %d to %s\n",
					i, fnamebuf);
			}
			fd = open(fnamebuf, openflag, 0644);
			if (fd < 0) {
				perror("open");
				ricoh_bye();
				goto fatalerr;
			}
			ricoh_getpict(i, fd);
			close(fd);
		}
		break;

	case GETPICT1:
		if (!n) {
			fprintf(stderr, "no picts at all\n");
			break;
		}
		if (startpict != endpict) {
			fprintf(stderr, "more than 1 picts specified\n");
			ricoh_bye();
			goto fatalerr;
		}
		if (verbose) {
			fprintf(stderr,
				"getting image %d to stdout\n", startpict);
		}
		ricoh_getpict(startpict, STDOUT_FILENO);
		break;

#if 0
	case GETSTATUS:
		ricoh_getstatus();
#endif

	case NOTHING:
		if (verbose) {
			fprintf(stderr, "nothing to be done\n");
		}
		break;

	default:
		fprintf(stderr,
			"internal error: invalid mode %d\n", mode);
		break;
	}

	ricoh_bye();

	return;

fatalerr:
	/* cleanup() will be called automatically */
	exit(1);
}

void
usage()
{
	fprintf(stderr,
		"usage:\n"
		"\tdc3play [options] port\n"
		"\tdc3play [options] -o port > foo.jpg\n"
		"\tdc3play [options] -g port\n");
}

void
cleanup()
{
	close(fd0);
	uucpunlock(n0);
}

void
main(argc, argv)
	int argc;
	char **argv;
{
	char *s;

	uid = getuid();
	euid = geteuid();
	gid = getgid();
	egid = getegid();
	useruid();

	useuucplock = 1;

	while (--argc > 0 && **++argv == '-') {
		for (s = argv[0] + 1; *s != '\0'; s++) {
			switch (*s) {

			/* controls */
			case 'b':
				baudrateflag = atoi(s + 1);
				while (*s)
					s++;
				s--;
				break;
			case 'd':
				dumpflag = 1;
				if (s[1]) {
					dumpmaxlen = atoi(s + 1);
					while (*s)
						s++;
					s--;
				}
				break;
			case 'D':
				debugflag = 1;
				break;
			case 'i':
				openflag = O_WRONLY | O_EXCL | O_CREAT;
				break;
#if 0
			case 't':
				thumbnail = 1;
				break;
#endif
			case 'u':
				useuucplock = 0;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'E':
				echobackrate = atoi(s + 1);
				if (echobackrate < 1)
					echobackrate = 1;
				while (*s)
					s++;
				s--;
				break;
			case 'N':
				if (!s[1]) {
					fprintf(stderr, "no string specified"
						"for filename template\n");
					exit(1);
				}
				nametemplate = &s[1];
				while (*s)
					s++;
				s--;
				break;

			/* picture range */
			case 'e':
				endpict = atoi(s + 1);
				while (*s)
					s++;
				s--;
				break;
			case 's':
				startpict = atoi(s + 1);
				while (*s)
					s++;
				s--;
				break;
			case 'n':
				startpict = endpict = atoi(s + 1);
				while (*s)
					s++;
				s--;
				break;
			case 'a':
				allpict = 1;
				break;

			/* mode */
			case 'g':
				mode = GETPICTS;
				break;
			case 'o':
				mode = GETPICT1;
				break;
			case 'r':
				mode = RESET;
				break;
			case 'X':
				mode = NOTHING;
				break;
#if 0
			case 'S':
				mode = GETSTATUS;
				break;
#endif

			default:
				fprintf(stderr, "unknown option -%c\n", *s);
				usage();
				exit(1);
			}
		}
	}

	if (1 < argc) {
		fprintf(stderr, "invalid argument\n");
		usage();
		exit(1);
	} else if (argc == 0) {
		fprintf(stderr, "no device specified\n");
		usage();
		exit(1);
	}

	n0 = canondev(argv[0]);
	if (!n0) {
		fprintf(stderr, "inappropriate device specified, "
			"or device permission error\n");
		exit(1);
	}

	if (uucplock(n0) < 0) {
		fprintf(stderr, "device %s in use\n", n0);
		exit(1);
	}

	daemonuid();
	fd0 = open(n0, O_RDWR|O_NDELAY);
	if (fd0 < 0) {
		perror("open");
		exit(1);
	}
	useruid();

	atexit(cleanup);

	mainloop();

	exit(0);
}
